#![allow(non_camel_case_types)]
#![allow(non_snake_case)]
#![allow(deref_nullptr)]
#![allow(clippy::all)]

include!(concat!(env!("OUT_DIR"), "/dm-bindings.rs"));
include!(concat!(env!("OUT_DIR"), "/sem-bindings.rs"));
